"""Handle Shell Environment."""
import os


def ensure_env_var(var_name: str) -> str:
    """Load an environment variable as string.

    Parameters
    ----------
    var_name : str
        env name

    Returns
    -------
    str
        the env var, empty string of empty

    Raises
    ------
    LookupError
        if the env variable doesnt exist
    """
    var = os.getenv(var_name)
    if not var:
        raise LookupError(f'Missing env Variable "{var_name}"')
    return str(var)



"""Csv Related Tools."""
import csv
import logging
import re
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Type, Union

import pandas as pd

LOGGER = logging.getLogger(__name__)


def pandas_read_csv(
    file_path: Union[str, Path],
    decimal_char: Optional[str] = "",
    file_encoding: Optional[str] = "utf-8",
    **kwargs: Any,
) -> pd.DataFrame:
    """Read CSV File to dataframe and sniffs delimiter from first line.

    Columns that are Suffixed with ":type:typestring"
    will be explicitely converted.
    Example name:type:str will convert the column to string,
    and then remove ":type:str" from the colname

    Parameters
    ----------
    file_path : Union[str,PathLike]
        csv filepath to import
    decimal_char : str, optional
        pandas read decimal char. If empty will use . or , if dialect is ;
    file_encoding: str, optional
        encoding to open file with ,by default utf-8
    **kwargs : Any
        passed down to pd.read_csv

    Returns
    -------
    pd.DataFrame
    """
    file_path = Path(file_path) if not isinstance(file_path, Path) else file_path

    dialect = csv.Sniffer().sniff(file_path.open("r", encoding=file_encoding).readline())
    LOGGER.debug("Sniffed Dialect with delim: '%s' from %s", dialect.delimiter, file_path)

    decimal = decimal_char or "," if dialect.delimiter == ";" else "."

    typecols = _csv_read_type_cols(file_path, dialect, decimal)
    # fmt: on
    if typecols:
        renamer = {}
        type_dict: Dict[str, Any] = {}
        for org_col, (new_col, col_type) in typecols.items():
            renamer[org_col] = new_col
            type_dict[org_col] = col_type
        LOGGER.debug("Reading CSV with specific types for: %s", type_dict)
        type_csv: pd.DataFrame = pd.read_csv(
            str(file_path.absolute()),
            dialect=dialect,
            dtype=type_dict,
            **kwargs,
        )  # type: ignore
        return type_csv.rename(renamer, axis=1)

    if "engine" not in kwargs:
        kwargs["engine"] = "python"
    return pd.read_csv(file_path, dialect=dialect, decimal=decimal, **kwargs)  # type: ignore


def _csv_read_type_cols(file_path: Path, dialect: Type[csv.Dialect], decimal: str) -> Dict[str, Tuple[str, str]]:
    csv_header = pd.read_csv(file_path, dialect=dialect, decimal=decimal, nrows=0)  # type: ignore
    type_rx = re.compile("(?P<col>.*):type:(?P<type>.*)")
    # fmt: off
    typecols: Dict[str, Tuple[str, str]] = {
        str(col): (str(match.group("col")), str(match.group("type")))
        for col in csv_header
        if (match := type_rx.match(str(col)))
    }

    return typecols



"""Module to Facilitate the Translation from Odoo Records to another Instance."""
import datetime
import logging
import re
from typing import Any, Dict, Iterable, List, Optional, Union

from odoorpc.models import Model

from .helpers.rpc import get_rec_meta, mapped_rpc

LOGGER = logging.getLogger(__name__)


def map_record_value(
    record_value: Any,
    record: Model,
    field: str,
    mapper: Dict[str, Any],
) -> Any:
    """Translate record value if its contained in mapper.

    Since Odoorpc returns many2one records as tuple (id,name) we are only interested in the id.
    Parameters
    ----------
    record_value : Any
        Value to Cleanup
    record : Model
        record from which the value was taken
    field : str
        field from wich the value was taken.
    mapper : Dict[str, Any]
        dict of {fiel_name:mapper}. where mapper = {source:target}.
        All fields with a mapper will get it's value swapped.
        Usually used to translate relational ids

    Returns
    -------
    Any
        Value i no mapper was found. Mapped value if mapper was found.
    """
    if record_value and get_rec_meta(record, field).type == "many2one":
        record_value = record_value[0]
    if mapper := mapper.get(field):  # type: ignore
        if isinstance(mapper, dict):
            if isinstance(record_value, list):
                return [mapper[v] for v in record_value]
            if hasattr(record_value, "id"):
                if len(record_value) > 1:
                    return [mapper[v] for v in record.ids]
                return mapper[record_value.id]
            return mapper[record_value]
    return record_value


def map_record_values(
    record: Model,
    mapper_dict: Dict[str, Union[str, Dict[Any, Any]]],
    ignore_map_errs: bool = False,
) -> Dict[str, Any]:
    """Return a Dict Value with Values mapped according to mapper_dict.

    Parameters
    ----------
    record : Model
        odoo_rpc component. All Keys in Mapper_dict will be read
    mapper_dict : Union[str, Dict[Any, Any]]
        dict of {"field_name":mapper}
        where mapper is a dict of {source:target}
    ignore_map_errs : bool, optional
        wether to just leave map errors empty, by default False

    Returns
    -------
    Dict[str,Any]
        Dict of fields:mapped_value

    Raises
    -----
    KeyError
        when value not found in mapper
    """
    record.ensure_one()
    record_values = record.read(list(mapper_dict.keys()))[0]
    del record_values["id"]
    for key, value in record_values.items():
        try:
            value = map_record_value(value, record, key, mapper_dict)
        except Exception as err:
            if ignore_map_errs:
                continue
            raise KeyError(f"Couldn't find {value} in mapper {key}. Record: {record}") from err

        record_values[key] = value
    return record_values


def format_domain_values(
    domain: Iterable[Union[str, Iterable[Any]]],
    record: Model,
    field_mappers: Dict[str, Union[str, Dict[Any, Any]]],
) -> Iterable[Union[str, Iterable[Any]]]:
    """Take in Domain with %()s string templates.

    Parameters
    ----------
    domain : Iterable[Union[str, Iterable[Any]]]
        Original Domain with %(val)s string formatting
    record : Model
        odoo record to get %()s values from
    field_mappers : Dict[str, Union[str, Dict[Any, Any]]]
        dict of {fiel_name:mapper}. where mapper = {source:target}.
        All fields with a mapper will get it's value swapped.
        Usually used to translate relational ids

    Returns
    -------
    Iterable[Union[str, Iterable[Any]]]
        Domain with values from record templated in and mapped according to field mappers
    """
    record.ensure_one()
    out_domain: List[Union[str, Iterable[Any]]] = []
    for subdomain in domain:
        if isinstance(subdomain, str):
            out_domain.append(subdomain)
            continue
        new_out = list(subdomain)

        if isinstance(new_out[2], str):
            if map_template := re.match(r"^%\((.*)\)s$", new_out[2]):
                map_key = map_template.groups()[0]
                mapped_val = mapped_rpc(record, map_key)
                if not mapped_val:
                    new_out[2] = False
                else:
                    try:
                        new_out[2] = map_record_value(mapped_val, record, map_key, field_mappers)
                    except KeyError:
                        new_out[2] = mapped_val
                    if (
                        isinstance(new_out[2], (list, tuple))
                        and len(new_out[2]) == 1
                        and not new_out[1] in ["in", "not in"]
                    ):
                        new_out[2] = new_out[2][0]
                    if isinstance(new_out[2], (datetime.date, datetime.datetime)):
                        new_out[2] = new_out[2].isoformat()
        out_domain.append(new_out)

    return out_domain


def transfer_records_and_map_ids(
    source_model: Model,
    target_model: Model,
    keep_fields: Dict[str, Union[str, Dict[int, int]]],
    match_domain: Iterable[Union[str, Iterable[Any]]],
    source_domain: Optional[Iterable[Union[str, Iterable[Any]]]] = None,
) -> Dict[int, int]:
    """Transfer Records from one odoo instance to another.

    Parameters
    ----------
    source_model : Model
        model where to get records
    target_model : Model
        model where to put records
    keep_fields : Dict[str, Union[str, Dict[int, int]]]
        dict of fields to translate. You can add mappers to translate relational fields
    match_domain : Iterable[Union[str, Iterable[Any]]]
        domain to use for equality matching. Use %(field_name)s syntax to substiture target values.
    source_domain : Optional[Iterable[Union[str, Iterable[Any]]]], optional
        domain to apply to source to not transfer all records, by default None

    Returns
    -------
    Dict[int, int]
        Mapper with new and old IDs
    """
    source_domain = source_domain or []
    source_records = source_model.browse(source_model.search(source_domain))
    mapper = {}
    for index, srec in enumerate(source_records, 1):
        equality_domain = format_domain_values(match_domain, srec, field_mappers=keep_fields)
        trec_id = target_model.search(equality_domain)
        if not trec_id:
            srec_vals = map_record_values(srec, keep_fields)
            for field, action in keep_fields.items():
                if action == "HTML":
                    if not srec_vals[field]:
                        srec_vals[field] = "<p><br></p>"
            trec_id = target_model.create(srec_vals)
            print(f"{index}/{len(source_records)} Created: {srec_vals.get('name') or srec_vals.get('id')}")
        if isinstance(trec_id, (list, tuple)) and len(trec_id) == 1:
            trec_id = trec_id[0]
        mapper[srec.id] = trec_id
    return mapper



"""Metaclass for Datasets."""
import logging
from pathlib import Path
from typing import Any, List, Optional

import pandas as pd

from godoo_rpc.helpers.pandas.read import pandas_read_csv

LOGGER = logging.getLogger(__name__)


class Dataset:
    """Class to bundle filesystem to dataframe."""

    def __init__(
        self,
        file: Path,
        dataframe: Optional[pd.DataFrame] = None,
        reference: str = "",
        read_only: bool = False,
        save_cols: Optional[List[str]] = None,
    ) -> None:
        """Construct Dataset.

        Parameters
        ----------
        file : Path
            FilePath
        dataframe : pd.DataFrame, optional
            Data, by default None
        reference : str, optional
            Reference for API, by default ""
        read_only : bool, optional
            prevent save method, by default False
        save_cols : List[str], optional
            list of columns to consider for saving, by defau
        """
        if isinstance(file, str):
            file = Path(file)
        self.file = file
        self._dataframe = dataframe
        self.reference = reference
        self._read_only = read_only
        self.save_cols = save_cols

    def _read(self, **kwargs: Any) -> pd.DataFrame:
        """Low Level Read method.

        Parameters
        ----------
        **kwargs : Any
            pass down to pandas.read_*

        Returns
        -------
        pd.DataFrame
            read dataframe

        Raises
        ------
        NotImplementedError
            no reader defined for file suffix
        ReferenceError
            File Not Found
        """
        if not self.file.exists():
            raise ReferenceError(f"File Not found: {self.file.absolute()}")
        if self.file.suffix == ".csv":
            return pandas_read_csv(self.file, **kwargs)
        if self.file.suffix == ".xlsx":
            return pd.read_excel(str(self.file.absolute()), **kwargs)
        if self.file.suffix == ".json":
            return pd.read_json(self.file, orient="table", precise_float=True, **kwargs)
        raise NotImplementedError(f"No Pandas load handler implemented for filetype {self.file.suffix}")

    def read(self, **kwargs: Any) -> pd.DataFrame:
        """Read the file to a Dataframe. Stores DF in this Class.

        Parameters
        ----------
        **kwargs : Any
            Pass down to _read

        Returns
        -------
        pd.DataFrame
            Content
        """
        LOGGER.info("Reading Dataset %s", self.file)

        if self._dataframe is None:
            self._dataframe = self._read(**kwargs)
        return self._dataframe

    @property
    def dataframe(self) -> pd.DataFrame:
        """Get the Dataframe. Reads file if Dataset is not set.

        Returns
        -------
        pd.DataFrame
        """
        if self._dataframe is None:
            self._dataframe = self.read()
        return self._dataframe

    @dataframe.setter
    def dataframe(self, dataframe: pd.DataFrame) -> None:
        """Setter for Dataframe.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Pandas Dataframe
        """
        self._dataframe = dataframe

    @property
    def save_dataframe(self) -> pd.DataFrame:
        """self.dataframe filtered by self.save_cols.

        Returns
        -------
        pd.DataFrame
            Dataframe for Save operation
        """
        my_df = self.dataframe
        if self.save_cols:
            return my_df[self.save_cols]
        return my_df

    def save(self, force_save: bool = False, **kwargs: Any) -> None:
        """Save Dataframe to Disk.

        Parameters
        ----------
        force_save : bool, optional
            wether to also save when there is no change
        **kwargs : Any
            Pass down to pandas.to_*

        Raises
        ------
        PermissionError
            File is marked Readonly
        NotImplementedError
            Unknown Fileformat
        """
        if self._read_only:
            raise PermissionError(f"Cant save '{self.file}'. Dataset is marked as Readonly.")
        save_df = self.save_dataframe
        if self.file.exists() and not force_save:
            if save_df.round(10).fillna("").equals(self._read().round(10).fillna("")):
                LOGGER.info(
                    "Skipping Save of Dataset with shape '%s' because the file '%s' didn't change",
                    self.dataframe.shape,
                    self.file,
                )
                return
        LOGGER.info("Saving Dataset with Shape %s to: %s", self.dataframe.shape, self.file)
        self.file.unlink(missing_ok=True)
        if self.file.suffix == ".csv":
            save_df.to_csv(self.file, index=False, **kwargs)
        elif self.file.suffix == ".xlsx":
            save_df.to_excel(self.file, index=False, **kwargs)
        elif self.file.suffix == ".json":
            save_df.to_json(self.file, orient="table", indent=4, **kwargs)
        else:
            raise NotImplementedError(f"No Save method for: {self.file.suffix}")



"""Provide Odoo CSV import Class and Methods."""

import logging
from pathlib import Path

from odoorpc import ODOO

from ..helpers.importer import OdooImporter
from ..helpers.pandas.read import pandas_read_csv
from .import_data import OdooDataImporter

LOGGER = logging.getLogger(__name__)


class OdooCsvImporter(OdooImporter):
    """Provide data pulling from a CSV File."""

    def __init__(self, session: ODOO) -> None:
        """Construct OdooCsvImporter.

        Parameters
        ----------
        session : ODOO
            Odoo Session
        """
        super().__init__(session=session)
        self.importer = OdooDataImporter(session=self.session)

    def upload(
        self,
        file_path: Path,
        model_name: str,
        max_batch_size: int = 0,
    ) -> None:
        """Upload Csv to Odoo.

        Parameters
        ----------
        file_path : Path
            Path to CSV File
        model_name : str
            Odoo Model Name
        max_batch_size : int
            Max Rows to sent in one Import action , by default len(dataframe)
        """
        self.importer.upload(
            dataframe=pandas_read_csv(file_path=file_path),
            model_name=model_name,
            max_batch_size=max_batch_size or 0,
            source_str=str(file_path),
        )



"""gOdoo-rpc."""
__version__ = "0.1.2"
from . import odoo2odoo_mapper  # noqa
from .api import OdooApiWrapper  # noqa
from .file_2_odoo import import_data, import_dataset  # noqa
from .login import login_odoo, login_odoo_env  # noqa
from .odoo_dataset import OdooDataset  # noqa



"""Module for res.config.settings importing."""

import logging

import pandas as pd
from odoorpc.models import Model

from ..helpers.importer import OdooImporter

LOGGER = logging.getLogger(__name__)


class OdooTranslationImporter(OdooImporter):
    """Odoo handler for res.config.settings."""

    def import_translations(self, dataframe: pd.DataFrame) -> None:
        """Import Translations by dataframe.

        Colnames need to be lang/fieldname. Example: en_US/name.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe with ID in first col
        """
        dataframe = dataframe.set_index("id")
        for index, (odoo_id, row) in enumerate(dataframe.iterrows(), 1):
            LOGGER.info("Importing Translations into Odoo (%s/%s) ", index, len(dataframe))
            odoo_item: Model = self.session.env.ref(odoo_id)
            if odoo_item:
                for col in dataframe.columns:
                    if pd.notna(row[col]):
                        col_info = col.split("/")
                        lang = col_info[0]
                        model_field = ".".join(col_info[1:])
                        odoo_item = odoo_item.with_context(lang=lang)
                        odoo_item.write({model_field: row[col]})



"""Provides the Odoo Session."""
import logging
from datetime import datetime, timedelta
from pathlib import Path
from time import sleep
from urllib.parse import urlparse

import odoorpc
import requests
from dotenv import load_dotenv

from .api import OdooApiWrapper
from .helpers.env import ensure_env_var

logger = logging.getLogger(__name__)


def wait_for_odoo(
    odoo_host: str, odoo_db: str, odoo_user: str, odoo_password: str, timeout_seconds: int = 600
) -> OdooApiWrapper:
    """Wait for Odoo Connection.

    Parameters
    ----------
    odoo_host : str
    odoo_db : str
    odoo_user : str
    odoo_password : str
    timeout_seconds : int, optional
        timeout to wait before raising, by default 600

    Returns
    -------
    OdooApiWrapper

    Raises
    ------
    TimeoutError
        If login timeout exeeded
    """
    start_time = datetime.now()
    while True:
        try:
            odoo_api = login_odoo(
                odoo_host=odoo_host,
                odoo_db=odoo_db,
                odoo_user=odoo_user,
                odoo_password=odoo_password,
            )
            return OdooApiWrapper(odoo_api)
        except (requests.HTTPError, requests.ConnectionError):
            sleep(1)
        if datetime.now() - timedelta(minutes=timeout_seconds) >= start_time:
            raise TimeoutError(f"Could not reach odoo after timeout of {timeout_seconds} seconds")


def login_odoo(
    odoo_host: str,
    odoo_db: str,
    odoo_user: str,
    odoo_password: str,
) -> odoorpc.ODOO:
    """Make sure ODOO RPC is connected and authenticated.

    Parameters
    ----------
    odoo_host : str
        Url to Odoo
    odoo_db : str
        odoo db name
    odoo_user : str
        login user
    odoo_password : str
        login password

    Returns
    -------
    odoorpc.ODOO
        odoo  rpc session

    Raises
    ------
    LookupError
        When db could not be found in Odoo
    """
    myurl = urlparse(odoo_host, allow_fragments=True)

    requests.head(myurl.geturl(), timeout=1200).raise_for_status()
    port = myurl.port or 80
    if port == 80 and myurl.scheme == "https":
        port = 443
    logger.info("Connecting to Odoo instance on: %s:%s", myurl.hostname, port)

    rpc_session = odoorpc.ODOO(
        host=myurl.hostname,
        port=port,
        timeout=300,
        protocol="jsonrpc+ssl" if myurl.scheme == "https" else "jsonrpc",
    )

    logger.debug(
        "Logging Into Odoo DB=%s, User=%s Password=%s",
        odoo_db,
        odoo_user,
        "*" * len(odoo_password),
    )
    rpc_session.login(odoo_db, odoo_user, odoo_password)
    return rpc_session


def login_odoo_env(dotenv_path: str = ".env", override_env: bool = False) -> odoorpc.ODOO:
    """Log into Odoo using Env Vars.

    Parameters
    ----------
    dotenv_path : str, optional
        Path to .env file, by default ".env"
    override_env : bool, optional
        Wether to override existing env variables with .env, by default False

    Returns
    -------
    OdooApiWrapper
        Logged in Odoo API
    """
    if Path(dotenv_path).exists():
        load_dotenv(dotenv_path, override=override_env)
    return login_odoo(
        odoo_host=ensure_env_var("ODOO_HOST"),
        odoo_db=ensure_env_var("ODOO_DB"),
        odoo_user=ensure_env_var("ODOO_USER"),
        odoo_password=ensure_env_var("ODOO_PASSWORD"),
    )



"""Importer classes for Odoo."""
from .import_csv import OdooCsvImporter  # noqa
from .import_data import OdooDataImporter  # noqa
from .import_product_images import OdooImageImporter  # noqa
from .import_settings import OdooSettingsImporter  # noqa
from .import_translation import OdooTranslationImporter  # noqa
from .import_xl import OdooXLImporter  # noqa



"""Helper Methods for Odoo."""

from odoorpc import ODOO


class OdooImporter:
    """Base Class for Odooimporters."""

    def __init__(self, session: ODOO) -> None:
        """Initialize OdooImporter.

        Parameters
        ----------
        session : ODOO
            Odoo Session
        """
        self.session = session



"""Wrapper scripts for OdooRPC."""
import logging
from pathlib import Path

from odoorpc import ODOO

from .importers import (
    OdooCsvImporter,
    OdooDataImporter,
    OdooImageImporter,
    OdooSettingsImporter,
    OdooTranslationImporter,
    OdooXLImporter,
)

logging.getLogger("odoorpc.rpc.jsonrpclib").setLevel(logging.INFO)


class OdooApiWrapper:
    """Wrap the Odoo RPC Api."""

    def __init__(self, odoo_session: ODOO) -> None:
        """Construct OdooApiWrapper.

        Parameters
        ----------
        odoo_session : ODOO
            Logged in Odoo session
        """
        self.session = odoo_session
        self.csv_importer = OdooCsvImporter(self.session)
        self.data_importer = OdooDataImporter(self.session)
        self.image_importer = OdooImageImporter(self.session)
        self.settings_importer = OdooSettingsImporter(self.session)
        self.translation_importer = OdooTranslationImporter(self.session)

    def xl_importer(self, file_path: Path, max_batch_size: int = 256) -> OdooXLImporter:
        """Return OdooXLImporter for the current session.

        Parameters
        ----------
        file_path : Path
            Path to the Excel file
        max_batch_size : int, optional
            max records to sent do oodoo in one request,, by default 256

        Returns
        -------
        OdooXLImporter
        """
        return OdooXLImporter(session=self.session, file_path=file_path, max_batch_size=max_batch_size)



"""Module for res.config.settings importing."""

import logging
from typing import Any, Dict, List

from odoorpc.models import Model

from ..helpers.importer import OdooImporter

LOGGER = logging.getLogger(__name__)


class OdooSettingsImporter(OdooImporter):
    """Odoo handler for res.config.settings."""

    def import_settings(self, settings: Dict[str, Any], lang: str = "") -> None:
        """Create Transient res.config.settings model and executes it.

        Parameters
        ----------
        settings : Dict[str, str]
            dict of setting:value
        lang: str
            language string to set context to
        """
        settings_env: Model = self.session.env["res.config.settings"]

        for setting, value in settings.items():  # Replace Ref Attributes with their internal ID
            if hasattr(settings_env._columns[setting], "relation"):  # pylint: disable=protected-access
                settings[setting] = self.session.env.ref(value).id

        if lang:
            LOGGER.info("Odoo preparing to set %s settings for language: %s", len(settings), lang)
            odoo_settings_id = settings_env.with_context(lang=lang).create([settings])
        else:
            LOGGER.info("Odoo preparing to set %s settings", len(settings))
            odoo_settings_id = settings_env.create([settings])

        LOGGER.info("Committing %s Settings to Odoo", len(settings))
        settings_env.execute(odoo_settings_id)  # Commit settings to Odoo

    def install_modules(self, modules: List[str]) -> None:
        """Install Odoo Modules by Name.

        Parameters
        ----------
        modules : List[str]
            List of Odoo module internal names
        """
        odoo_module_env: Model = self.session.env["ir.module.module"]
        odoo_module_env.update_list()
        LOGGER.info("Installing %s Odoo Modules", len(modules))
        for index, module in enumerate(modules, 1):
            module_id = odoo_module_env.search([("state", "!=", "installed"), ("name", "=", module)])
            if module_id:
                LOGGER.info("Odoo Installing Module (%s/%s): %s", index, len(modules), module)
                odoo_module_env.browse(module_id).button_immediate_install()



"""Provide Odoo Excel import Class and Methods."""


import logging
import re
from pathlib import Path
from typing import Any, Dict, Optional

import pandas as pd
from odoorpc import ODOO
from openpyxl import Workbook, load_workbook

from ..helpers.importer import OdooImporter
from .import_data import OdooDataImporter
from .import_settings import OdooSettingsImporter
from .import_translation import OdooTranslationImporter

LOGGER = logging.getLogger(__name__)


class OdooXLImporter(OdooImporter):
    """Provide data pulling from a Excel File."""

    def __init__(self, session: ODOO, file_path: Path, max_batch_size: int = 256) -> None:
        """Initialize OdooXLImporter.

        Parameters
        ----------
        session : ODOO
            Odoo Session
        file_path : Path
            Excel FilePath
        max_batch_size : int, optional
            Max num of recrords to send to Odoo at once, by default 256
        """
        self.path = file_path
        self._workbook: Optional[Workbook] = None
        self.max_batch_size = max_batch_size
        super().__init__(session=session)

    @property
    def workbook(self) -> Workbook:
        """Openpyxl workbook. Cached.

        Returns
        -------
        Workbook
            OpenpyXL Workbook

        Raises
        ------
        FileNotFoundError
            Couldn't find/open workbook
        """
        if not self._workbook:
            self._workbook = load_workbook(str(self.path), read_only=True)
        if not self._workbook:
            raise FileNotFoundError(f"Workbook {self.path} Not Loaded")
        return self._workbook

    def import_sheets_by_regex(self, regex_pattern: str) -> None:
        """Import all Worksheets with name matching regex into the odoo module defined in the regex.

        Parameters
        ----------
        regex_pattern : str
            Regex pattern with named group "module"
        """
        worksheets = self.workbook.worksheets

        import_sheets = {match: s for s in worksheets if (match := re.search(regex_pattern, str(s.title)))}
        importer = OdooDataImporter(session=self.session)
        for index, (match, sheet) in enumerate(import_sheets.items(), 1):
            odoo_model = match.group("model")
            LOGGER.info(
                "(%s/%s) Importing Worksheet: %s into Module %s", index, len(import_sheets), sheet.title, odoo_model
            )
            sheet_data = pd.read_excel(str(self.path.resolve()), sheet.title)

            importer.upload(
                source_str=str(self.path.stem) + str("!") + str(sheet.title),
                model_name=odoo_model,
                dataframe=sheet_data,
                max_batch_size=self.max_batch_size,
            )

    def import_settings(self, sheet_name: str) -> None:
        """Import Odoo Settings from Excel Sheet Table.

        Parameters
        ----------
        sheet_name : str
            Sheet to import
        """
        settings_importer = OdooSettingsImporter(self.session)
        sheet_data: pd.DataFrame = pd.read_excel(str(self.path.resolve()), sheet_name=sheet_name)

        sheet_data = sheet_data.fillna("")
        for lang, group in sheet_data.groupby("Language"):
            lang_settings: Dict[str, Any] = pd.Series(
                group["Value"].values, index=group["Setting"].values  # type: ignore
            ).to_dict()
            settings_importer.import_settings(lang_settings, str(lang))

    def install_modules(self, sheet_name: str) -> None:
        """Install Modules Provided by List.

        Parameters
        ----------
        sheet_name : str
            Sheetname with "Name" column
        """
        settings_importer = OdooSettingsImporter(self.session)
        sheet_data: pd.DataFrame = pd.read_excel(str(self.path.resolve()), sheet_name=sheet_name)
        settings_importer.install_modules(sheet_data["Name"].tolist())

    def import_translations(self, sheet_name: str) -> None:
        """Import Translations from Excel sheet.

        Parameters
        ----------
        sheet_name : str
            excel sheetname
        """
        translation_importer = OdooTranslationImporter(self.session)
        sheet_data: pd.DataFrame = pd.read_excel(str(self.path.resolve()), sheet_name=sheet_name)
        translation_importer.import_translations(sheet_data)



"""Provide Odoo CSV import Class and Methods."""

import csv
import io
import json
import logging
import re
from dataclasses import dataclass
from typing import Any, Dict, List, Match

import numpy as np
import pandas as pd
from odoorpc import ODOO
from odoorpc.models import Model

from ..helpers.importer import OdooImporter

LOGGER = logging.getLogger(__name__)
LANG_COL_REGEX = re.compile(r"(?P<col>.*):lang:(?P<lang>.*)")


@dataclass
class LangPair:
    """Store a col -> languagecol relation."""

    id_col: str  # DF Col for Odoo affected External ID
    val_col: str  # DF Col for Lang Value
    field_name: str  # Odoo FieldName
    lang: str  # Odoo Context Language


def dataframe_get_lang_cols(dataframe: pd.DataFrame) -> Dict[str, Match[str]]:
    """Get Language Cols in Upload Dataset (fieldname:lang:en_US).

    Parameters
    ----------
    dataframe : pd.DataFrame
        dataframe with col names in sytle: fieldname:lang:en_US

    Returns
    -------
    Dict[str, re.Match[str]]
        dict of colname and re.match
    """
    return {col: re_match for col in dataframe.columns if (re_match := LANG_COL_REGEX.match(col))}


def dataframe_strip_language(dataframe: pd.DataFrame) -> pd.DataFrame:
    """Remove Language cols (see dataframe_get_lang_cols) from Dataframe.

    Parameters
    ----------
    dataframe : pd.DataFrame
        dataframe to remove cols from

    Returns
    -------
    pd.DataFrame
    """
    lang_cols = dataframe_get_lang_cols(dataframe=dataframe)
    other_cols = [col for col in dataframe.columns if col not in lang_cols]
    return dataframe[other_cols]


def dataframe_generate_lang_pairs(dataframe: pd.DataFrame) -> List[LangPair]:
    """Generate Language Pairs from Dataframe Columns.

        Each Pair Links the FieldCol to the Language col and the affected ID col.

    Parameters
    ----------
    dataframe : pd.DataFrame
        dataframe to scan for language cols

    Returns
    -------
    List[LangPair]
    """
    lang_cols = dataframe_get_lang_cols(dataframe=dataframe)
    lang_pairs: List[LangPair] = []
    for col, match in lang_cols.items():
        lang = str(match.group("lang"))
        ref_col = str(match.group("col"))
        id_col = ""
        if "/" in ref_col:
            splits = ref_col.split("/")
            try_index = 1
            while not id_col:
                id_try = "/".join(splits[:-try_index] + ["id"])
                if id_try in dataframe.columns:
                    id_col = id_try
                    ref_col = ".".join(splits[-try_index:])
        else:
            id_col = "id"
        lang_pairs.append(LangPair(id_col=id_col, val_col=col, field_name=ref_col, lang=lang))
    return lang_pairs


def chunk_odoo_import(full_data: pd.DataFrame, chunk_size: int) -> List[pd.DataFrame]:
    """Split Dataframe into chunks of max length. Can overflow a chunk until "id" column is not nan.

    Parameters
    ----------
    full_data : pd.DataFrame
        input dataframe
    chunk_size : int
        cutoff chunk size

    Returns
    -------
    List[pd.DataFrame]
        chunked dataframe
    """
    if len(full_data) <= chunk_size:
        return [full_data]

    LOGGER.debug("Chunking Dataset with %s Entries to size %s", len(full_data), chunk_size)
    chunk_count = -(-len(full_data) // chunk_size)  # Round Number of slices up
    chunks: List[pd.DataFrame] = np.array_split(full_data, chunk_count)  # type: ignore
    out_chunks: List[pd.DataFrame] = [chunks.pop(0)]  # Remove First Chunk and add to Output
    for index, chunk in enumerate(chunks):

        empty_ids = pd.isnull(chunk.id)  # Check Empty Ids

        if empty_ids.all():
            # No Ids at all? Append chunk to last Output chunk
            out_chunks[-1] = out_chunks[-1].append(chunk)  # type: ignore
            continue

        if not empty_ids.any():
            # No Empty Ids.. Just add Chunk to outlist
            out_chunks.append(chunk)
            continue

        move_rows_count = np.where(~empty_ids)[0][0]  # Get index of first Non empty Id.

        LOGGER.debug("(%s/%s) Move %s rows to Previous Chunk", index + 1, len(chunks), move_rows_count)
        move_rows = chunk.iloc[0:move_rows_count]
        out_chunks[-1] = out_chunks[-1].append(move_rows)  # type: ignore
        if not (out_chunk := chunk.drop(list(move_rows.index))).empty:
            out_chunks.append(out_chunk)

    LOGGER.info("Dataframe with len %s chunked into %s parts", len(full_data), len(out_chunks))
    return out_chunks


class OdooDataImporter(OdooImporter):
    """Provide data pulling from a CSV File.

    Dataframe needs to be in standard Odoo Format.

    In addition there are some extra features:

    Suffix a column with ":lang:langCoge" example: "name:lang:en_US"
        This will add Translations in the sepecified language
    """

    def __init__(
        self,
        session: ODOO,
    ) -> None:
        """Construct OdooDataImporter.

        Parameters
        ----------
        session : ODOO
            Odoo Session

        """
        self._source = ""
        self._model_name = ""
        self._max_batch_size = 0
        self._dialect = csv.excel
        super().__init__(session=session)

    @property
    def model_name(self) -> str:
        """Expose Model_Name as ReadOnly Property.

        Returns
        -------
        str
            Odoo Model Name
        """
        return self._model_name

    def _df2upload_file(self, data: pd.DataFrame) -> str:
        """Convert Dataframe into OdooRPC Upload String.

        Parameters
        ----------
        data : pd.DataFrame
            Model data

        Returns
        -------
        str
            CSV String
        """
        csv_sio = io.StringIO()
        data.to_csv(
            csv_sio,
            encoding="utf-8",
            sep=self._dialect.delimiter,
            date_format="%Y-%m-%d %H:%M:%S",
            index=False,
            quotechar=self._dialect.quotechar or "",
            line_terminator=self._dialect.lineterminator,
            escapechar=self._dialect.escapechar,
        )  # Convert back to in memory csv
        return csv_sio.getvalue()

    def _import_translations(self, dataframe: pd.DataFrame) -> None:
        lang_pairs = dataframe_generate_lang_pairs(dataframe=dataframe)
        if not lang_pairs:
            return
        LOGGER.info("Found: %s Language Columns", len(lang_pairs))
        for index, pair in enumerate(lang_pairs, 1):
            LOGGER.info("(%s/%s) Processing Language %s, Col: %s", index, len(lang_pairs), pair.lang, pair.val_col)
            for _, row in dataframe.iterrows():
                if pd.notna(row[pair.val_col]):
                    odoo_rec: Model = self.session.env.ref(row[pair.id_col])
                    odoo_rec = odoo_rec.with_context(lang=pair.lang).write({pair.field_name: row[pair.val_col]})

    def _strip_existing_records(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        """Remove Records, form Dataframe if the id is already registered in odoo ir.model.data.

        Parameters
        ----------
        dataframe : pd.DataFrame
            input dataframe with "id" column

        Returns
        -------
        pd.DataFrame
            filtered dataset
        """
        xml_ids = dataframe.id.dropna().drop_duplicates().str.split(".", n=1, expand=True)
        odoo_model_data: Model = self.session.env["ir.model.data"]
        existing_ids = odoo_model_data.search_read(
            [
                "&",
                ("name", "in", list(xml_ids[1])),
                ("module", "in", list(xml_ids[0])),
            ],
            fields=["name", "module"],
        )
        if not existing_ids:
            return dataframe
        odoo_xml_id_df = pd.DataFrame(existing_ids)
        odoo_xml_ids = odoo_xml_id_df["module"] + "." + odoo_xml_id_df["name"]

        return dataframe[~dataframe.id.ffill().isin(odoo_xml_ids)]

    def upload(
        self,
        dataframe: pd.DataFrame,
        model_name: str,
        max_batch_size: int = 0,
        source_str: str = "Internal",
        skip_existing_ids: bool = False,
    ) -> None:
        """Upload Dataframe to Odoo.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe to upload. Must contain "id" column
        model_name : str
            Odoo Model Name
        max_batch_size : int
            Max Rows to sent in one Import action , by default len(dataframe)
        source_str : str, optional
            Just for Logging, datasource, by default "Internal"
        skip_existing_ids : bool, optional
            Wether to first query Odoo for Ids and skip them if neccessary

        Raises
        ------
        IndexError
            When there are Duplicate IDs in the Dataframe
        """
        self._source = source_str or ""
        self._model_name = model_name
        self._max_batch_size = max_batch_size if max_batch_size > 0 else len(dataframe)

        LOGGER.info("Staring Import from '%s' to '%s'", self._source, self._model_name)
        if dataframe.empty:
            LOGGER.error("Cant import %s. No Data Provided", self._source)
            return

        if (dupe_mask := dataframe["id"].replace("", np.nan).dropna().duplicated()).any():
            dupes = dataframe.dropna(subset=["id"]).loc[dupe_mask, "id"]
            LOGGER.error("Detetced duplicate Ids in %s. Dupes: %s", self._source, dupes)
            raise IndexError(f"Detedced duplicate Ids in {self._source}. Dupes:\n{dupes}")

        if skip_existing_ids:
            dataframe = self._strip_existing_records(dataframe)
            if dataframe.empty:
                LOGGER.info("All IDs in Dataframe already exist in Odoo")
                return

        dataframe = dataframe.loc[:, ~dataframe.columns.str.contains("^Unnamed")]

        self._upload_chunked(dataframe)

    def _upload_chunked(self, dataframe: pd.DataFrame) -> None:
        """Upload the Specified Dataframe chunked.

        Parameters
        ----------
        dataframe : pd.DataFrame
            dataframe to upload
        """
        no_lang_df = dataframe_strip_language(dataframe=dataframe)
        no_empty_rows = no_lang_df.dropna(how="all", axis=0)
        chunks = chunk_odoo_import(no_empty_rows, self._max_batch_size)
        for index, chunk in enumerate(chunks, 1):
            LOGGER.info(
                "(%s/%s) Importing %s records from '%s' into Odoo", index, len(chunks), len(chunk), self._source
            )
            self._upload(chunk, index)
        self._import_translations(dataframe)

    def _upload(self, dataset: pd.DataFrame, index: int = 1) -> None:
        """Create a base_import transient model and commits provided data it to odoo.

        Parameters
        ----------
        dataset : pd.DataFrame
            Dataset to upload
        index : int, optional
            folename index for logging in case of chunked upload, by default 1
        """
        odoo_model: Model = self.session.env["base_import.import"]
        up_data = self._df2upload_file(dataset)

        imp_id = odoo_model.create(
            {
                "res_model": self._model_name,
                "file": up_data,
                "file_type": "text/csv",
                "file_name": self._source + f"-{index}",
            }
        )
        LOGGER.debug("Odoo: Created 'base.import' model with id : %s", imp_id)
        imp = odoo_model.browse(imp_id)

        headers = list(dataset.columns.values)

        resp = imp.do(
            headers,
            headers,
            {
                "headers": True,
                "advanced": True,
                "keep_matches": False,
                "date_format": "%Y-%m-%d %H:%M:%S",
                "datetime_format": "%Y-%m-%d %H:%M:%S",
                "encoding": "utf-8",
                "separator": self._dialect.delimiter,
                "quoting": self._dialect.quotechar,
                "float_thousand_separator": ",",
                "float_decimal_separator": ".",
            },
        )

        if messages := resp["messages"]:
            self.handle_upload_errors(messages, dataset)

        expect_len = len(dataset["id"].dropna().unique())
        if len(resp["ids"]) != expect_len:
            LOGGER.error("Expected %s, Records to be imported, but odoo only reports %s", expect_len, len(resp["ids"]))

    def handle_upload_errors(self, messages: List[Dict[str, Any]], dataset: pd.DataFrame) -> None:
        """Handle Error Messages.

        Parameters
        ----------
        messages : List[Dict[str, Any]]
            Response from Upload
        dataset : pd.DataFrame
            The Reference Dataset

        Raises
        ------
        ImportError
            If an error Occured
        """
        LOGGER.error("Odoo Import Failed with message:\n%s", json.dumps(messages, sort_keys=True, indent=2))
        messages = [m for m in messages if "rows" in m]
        affected_row_pairs = [(int(m["rows"]["from"]), int(m["rows"]["to"])) for m in messages]
        affected_rows = [row for lower, upper in affected_row_pairs for row in range(lower, upper + 1)]
        err_df = dataset.iloc[list(set(affected_rows))]

        affected_fields: List[str] = [str(m.get("field")) for m in messages] + ["id"]
        affected_cols: List[str] = []
        for col in affected_fields:  # Need to add /id or /name if we are talking about a relational field
            if col in err_df.columns:
                affected_cols.append(col)
            elif col + "/id" in err_df.columns:
                affected_cols.append(col + "/id")
            elif col + "/name" in err_df.columns:
                affected_cols.append(col + "/name")

        err_df = err_df[[c for c in err_df.columns if c in affected_cols]]  # Looks weird, but preserves DF Col Order
        with pd.option_context("display.max_rows", None, "display.max_columns", None):  # type: ignore
            LOGGER.error("Relevant Dataset:\n%s", err_df)
        raise ImportError("Odoo Responded with errors. See log.")



"""Imports Data from Filesystem Tables into Odoo.

Files should be named in Schema:

001_Base/001_odoo.module.csv

Import order is Alphanumeric.
"""

import logging
import re
from pathlib import Path
from typing import List, Pattern

from .api import OdooApiWrapper
from .odoo_dataset import OdooDataset, import_dataset, import_dataset_timestamped

FOLDER_REGEX = re.compile(r"(\d+)_.*")

_logger = logging.getLogger(__name__)


def gather_import_files(datafolder: Path, regex_pattern: Pattern[str]) -> List[OdooDataset]:
    """Gather OdooDatasets by Regex.

    Parameters
    ----------
    datafolder : Path
        Folder to recursively search in
    regex_pattern : re.Pattern[str]
        regex pattern to match files on. Needs group "Module"

    Returns
    -------
    List[OdooDataset]
        List of Datasets
    """
    out_list = []
    for file in datafolder.rglob("*"):
        if match := regex_pattern.match(file.name):
            out_list.append(OdooDataset(file=file, reference=match.group("module")))

    out_list = sorted(out_list, key=lambda f: str(f.sort_key(datafolder)))

    return out_list


def import_data(
    read_path: Path,
    data_regex: str,
    product_image_regex: str,
    odoo_api: OdooApiWrapper,
    skip_existing_ids: bool = False,
    check_dataset_timestamp: bool = False,
) -> None:
    r"""Import Files found in folder.

    Files should be named index_odoo.module.csv.
    Folders can also be prepended with index_.
    Files will be sorted as a combination of all indices.

    Parameters
    ----------
    read_path : Path
        Folder to search for files or file to import
    data_regex : re.Pattern[str]
        Regex for Datafiles. Needs Group "Module"
    product_image_regex : re.Pattern[str]
        Regex for Image files. Example: (?P<default_code>\d{6})\.(jpeg|png|jpg)
    odoo_api : OdooApiWrapper
        logged in Odoo Api Instance
    skip_existing_ids : bool, optional
        Skip Ids that already exist in Odoo, by default False
    check_dataset_timestamp : bool, optional
        Log Import timestamps in ir.config_parameters. Skips Unchanged Datasets, by default false

    Raises
    ------
    NameError
        When the filename couldnt be parsed
    """
    data_regex_comp = re.compile(data_regex)

    if read_path.is_dir():
        import_files = gather_import_files(datafolder=read_path, regex_pattern=data_regex_comp)
    elif read_path.suffix == ".py":
        import_files = [OdooDataset(file=read_path, reference=read_path.stem)]
    elif match := data_regex_comp.match(read_path.name):
        import_files = [OdooDataset(file=read_path, reference=match.group("module"))]
    else:
        raise NameError(f"Couldnt Parse: {read_path.name}")
    _logger.info("Collected: %s files", len(import_files))
    max_index = len(import_files)
    for index, data in enumerate(import_files, 1):
        _logger.info("Processing Dataset (%s/%s) --> %s", index, max_index, data.reference)
        if check_dataset_timestamp:
            import_dataset_timestamped(
                dataset=data, odoo_api=odoo_api, skip_existing_ids=skip_existing_ids, relative_folder=read_path
            )
        else:
            import_dataset(data=data, odoo_api=odoo_api, skip_existing_ids=skip_existing_ids)

    if product_image_regex:
        product_image_regex_comp = re.compile(product_image_regex)

        image_files = odoo_api.image_importer.search_images_by_regex(
            read_path / "img", product_image_regex_comp.pattern
        )
        odoo_api.image_importer.import_product_images(image_files, overwrite_images=not skip_existing_ids)



"""Helper Methods for Odoo."""
from typing import Any, Callable, Collection, Literal, Union

from odoorpc.models import Model


def build_odoo_domain_from_tuples(
    tup: Collection[Collection[Any]], operator: Literal["|", "&"] = "&"
) -> Collection[Any]:
    """Flatten a list of tuples into a valid Domain string, with prepended operators.

    Parameters
    ----------
    tup : Collection[Collection[Any]]
        List like [(name,ilike,test),(model,=,product.product),...]
    operator : Literal["|", "&"], optional
        & or | , by default "&"

    Returns
    -------
    Collection[Any]
        Flattened list like this:
        [|,(name,ilike,test),(model,=,product.product)]

    Raises
    ------
    AssertionError
        If Operator is not any of | or &
    """
    if operator not in ["|", "&"]:
        raise AssertionError(f"Odoo Domain Argument must be | or &, not {operator}")

    operators = [*(operator * (len(tup) - 1))]
    return tuple(operators + operators)


def mapped_rpc(record: Model, field_accessor: Union[str, Callable[[Model], Any]]) -> Any:
    """Like ORM record.map, but working locally using native RPC calls.

    Parameters
    ----------
    record : Model
        Odoo record
    field_accessor : Union[str, Callable[[Model], Any]]
        string like "partner_id.name" or callable.
        Just like ORM Mapped.

    Returns
    -------
    Any
        Mapped result
    """
    if callable(field_accessor):
        return field_accessor(record)
    if "." in field_accessor:
        first_field = field_accessor.split(".")[0]
        rec = record[first_field]
        return mapped_rpc(rec, field_accessor.replace(first_field, "").strip("."))
    return record[field_accessor]


def get_rec_meta(record: Model, field_name: str) -> Any:
    """Get Column metadata by mapstring.

    Parameters
    ----------
    record : Model
        Odoo Record
    field_name : str
        fieldname or field accessor like "partner_id.name"

    Returns
    -------
    Any
        columns metadata of field
    """
    if "." in field_name:
        first_field = field_name.split(".")[0]
        rec = record[first_field]
        return get_rec_meta(rec, field_name.replace(first_field, "").strip("."))
    return record._columns[field_name]  # pylint: disable=protected-access



"""Facilitates Odoo Odoo FileImports."""

import importlib.util
import json
import logging
import re
import sys
from datetime import datetime
from importlib.abc import Loader
from pathlib import Path
from typing import Any, Dict

import pandas as pd

from .api import OdooApiWrapper
from .helpers.pandas.dataset import Dataset

FOLDER_REGEX = re.compile(r"(\d+)_.*")

_logger = logging.getLogger(__name__)


class OdooDataset(Dataset):
    """Class to Map Dataset to File and Odoo_Model."""

    def __init__(
        self,
        file: Path,
        dataframe: pd.DataFrame = None,
        reference: str = "",
        read_only: bool = False,
    ) -> None:
        """Construct OdooDataset.

        Parameters
        ----------
        file : Path
            FilePath
        dataframe : pd.DataFrame, optional
            Data, by default None
        reference : str, optional
            Reference for API, by default ""
        read_only : bool, optional
            prevent save method
        """
        self.is_script = False
        if file:
            if file.suffix == ".py":
                read_only = True
                self.is_script = True
        super().__init__(file, dataframe=dataframe, reference=reference, read_only=read_only)

    def sort_key(self, max_up_folder: Path) -> str:
        """Traverse up the Path Tree to build sortable path.

        Parameters
        ----------
        max_up_folder : Path
            stop at partent

        Returns
        -------
        str
            sort string
        """
        sort_list = [self.file.name]
        for par in self.file.parents:
            if FOLDER_REGEX.match(par.name):
                sort_list.append(par.name)
            if par == max_up_folder:
                break
        return "/".join(reversed(sort_list))


def import_settings_from_df(odoo_api: OdooApiWrapper, dataframe: pd.DataFrame) -> None:
    """Load Odoo Settings from col "Value" into setting from Col "Setting".

    Values with a "Language" specified will be loaded into the Language

    Parameters
    ----------
    odoo_api : OdooApiWrapper
        Odoo Api wrapper
    dataframe : pd.DataFrame
        Settings DF. Columns: "Setting,Value,Language"
    """
    dataframe.Language = dataframe.Language.fillna(False)
    for lang, group in dataframe.groupby("Language", sort=False):
        lang_settings: Dict[str, Any] = pd.Series(
            group["Value"].values, index=group["Setting"].values  # type: ignore
        ).to_dict()

        odoo_api.settings_importer.import_settings(lang_settings, str(lang) if not pd.isna(lang) else "")


def import_by_script(odoo_api: OdooApiWrapper, dataset: OdooDataset) -> None:
    """Run Method Main(odoo_api) in Python script.

    Parameters
    ----------
    odoo_api : OdooApiWrapper
        Odoo Api Wrapper
    dataset : OdooDataset
        Odoo Dataset CLass

    Raises
    ------
    ModuleNotFoundError
        Could not import Module
    ImportError
        Could not load Spec
    """
    _logger.debug("Attempting to load python Module: %s", dataset.file)
    spec = importlib.util.spec_from_file_location("odoo_data_import", dataset.file.absolute())
    if not spec:
        raise ModuleNotFoundError(f"Couldnt Load '{dataset.file.absolute()}' as module")
    mod = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = mod
    if not isinstance(spec.loader, Loader):
        raise ImportError(f"Couldnt Load Spec from '{dataset.file.absolute()}'")
    spec.loader.exec_module(mod)
    _logger.info("Launching Main Method in Import Script: %s", dataset.file)
    mod.Main(odoo_api)  # type: ignore


def import_model_data(
    odoo_api: OdooApiWrapper,
    data: OdooDataset,
    skip_existing_ids: bool = False,
) -> None:
    """Import Data into Odoo using base.import.

    Parameters
    ----------
    odoo_api : OdooApiWrapper
        Odoo API Wrapper
    data : OdooDataset
        OdooDataset
    skip_existing_ids : bool, optional
        Skip Ids that already exist in Odoo, by default False
    """
    if data.is_script:
        import_by_script(odoo_api=odoo_api, dataset=data)
        return

    odoo_api.data_importer.upload(
        data.dataframe, data.reference, 256, data.file.name, skip_existing_ids=skip_existing_ids
    )


def import_dataset(data: OdooDataset, odoo_api: OdooApiWrapper, skip_existing_ids: bool = False) -> None:
    """Import Dataset into Odoo Modules.

    Parameters
    ----------
    data : OdooDataset
        Import Dataset for Odoo Model or Archive
    odoo_api : OdooApiWrapper
        Odoo Api
    skip_existing_ids : bool, optional
        Skip Ids that already exist in Odoo, by default False

    Special Odoo "Modules":
        odoo-modules:
            all modules specified in col "Name" will be installed
        odoo-settings:
            Sets res.config.settings
            settings from col "Setting" will be set to col "Value"
            Optional col "Language" can specify the Lanuage to set.
        odoo-translate:
            translates Odoo Fields
            Col pattern is language/fieldname. Example: en_US/name
            'id' must be the first column

    Special Actions for Modules:
        stock.inventory:
            All imported inventories will get action_start and action_validate called after import.
    """
    if data.reference == "odoo-modules":
        _logger.info("Installing modules from '%s' into Odoo", data.file.name)
        odoo_api.settings_importer.install_modules(data.dataframe.Name.unique())

    elif data.reference == "odoo-settings":
        _logger.info("Importing Settings from '%s' into Odoo", data.file.name)
        import_settings_from_df(odoo_api=odoo_api, dataframe=data.dataframe)

    elif data.reference == "odoo-translate":
        _logger.info("Importing Translation from '%s' into Odoo", data.file.name)
        odoo_api.translation_importer.import_translations(data.dataframe)

    elif data.reference == "odoo-unarchive":
        unarch_ids = data.dataframe.id.unique()
        _logger.info("Unarchiving Items %s from '%s'", len(unarch_ids), data.file.name)
        for xml_id in unarch_ids:
            if not (odoo_rec := odoo_api.session.env.ref(xml_id)).active:
                _logger.debug("Unarchiving '%s' --> '%s'", xml_id, odoo_rec.name)
                odoo_rec.action_unarchive()

    elif data.reference == "odoo-archive":
        unarch_ids = data.dataframe.id.unique()
        _logger.info("Archiving Items %s from '%s'", len(unarch_ids), data.file.name)
        for xml_id in unarch_ids:
            if not (odoo_rec := odoo_api.session.env.ref(xml_id)).active:
                _logger.debug("Archiving '%s' --> '%s'", xml_id, odoo_rec.name)
                odoo_rec.action_archive()

    else:
        _logger.info("Importing '%s' into Odoo module '%s'", data.file.name, data.reference)
        import_model_data(odoo_api=odoo_api, data=data, skip_existing_ids=skip_existing_ids)


def import_dataset_timestamped(
    dataset: OdooDataset, odoo_api: OdooApiWrapper, relative_folder: Path, skip_existing_ids: bool = False
) -> None:
    """Call import_dataset, log Timestamp in ir.config.parameter and skip if timestamp is smaller.

    Parameters
    ----------
    dataset : OdooDataset
        Upload Datase
    odoo_api : OdooApiWrapper
        Logged in API Wrapper
    relative_folder : Path
        Folder to which odoo dataset paths are relative to
    skip_existing_ids : bool, optional
        Skip Ids that already exist in Odoo, by default False
    """
    paramteter_model = odoo_api.session.env["ir.config_parameter"]

    continue_import = True
    import_ref = dataset.file.relative_to(relative_folder)
    odoo_ref_name = "godoo_rpc_import_cache"
    change_date_fs = datetime.fromtimestamp(dataset.file.stat().st_mtime)

    import_ref_search = paramteter_model.search([("key", "=", odoo_ref_name)])
    odoo_import_ref = paramteter_model.browse(import_ref_search[0]) if import_ref_search else None

    change_dict = {}
    if odoo_import_ref:
        change_dict = json.loads(odoo_import_ref.value)
        if ref_entry := change_dict.get(str(import_ref)):
            odoo_change_date = datetime.fromisoformat(ref_entry)
            if change_date_fs <= odoo_change_date:
                _logger.debug("Skipping Import file because of Timestamp in Odoo: %s", dataset.file.absolute())
                continue_import = False

    if continue_import:
        import_dataset(data=dataset, odoo_api=odoo_api, skip_existing_ids=skip_existing_ids)
        change_dict.update({str(import_ref): change_date_fs.isoformat()})
        if odoo_import_ref:
            odoo_import_ref.value = json.dumps(change_dict)
        else:
            paramteter_model.create({"key": odoo_ref_name, "value": json.dumps(change_dict)})



"""Odoo Import API."""
import base64
import logging
import re
from pathlib import Path
from typing import Dict, List, Match, Tuple

from odoorpc.error import RPCError
from odoorpc.models import Model

from ..helpers.importer import OdooImporter

LOGGER = logging.getLogger(__name__)


class OdooImageImporter(OdooImporter):
    """Import images by regex into odoo."""

    def import_product_images(self, images: List[Tuple[str, Path]], overwrite_images: bool = False) -> None:
        """Search folder recursive for images matching a regex and then uploady matching ones to the odoo products.

        Parameters
        ----------
        images : List[tuple[str, Path]]
            mapping odoo default_code to image file
        overwrite_images : bool
            Wether Existing images should be overwritten
        """
        if not images:
            LOGGER.debug("Skipping Product image Import. No Images Provided")
            return

        product_ids = tuple(v[0] for v in images)
        try:
            products_model: Model = self.session.env["product.product"]
        except RPCError:
            LOGGER.warning("Cannot import Product images. Model product.product not found")
            return

        LOGGER.info("Querying Odoo with %s product IDs", len(product_ids))

        prod_ids = products_model.search([("default_code", "in", product_ids)])
        if not overwrite_images:
            existing_images = self.session.env["ir.attachment"].search_read(
                [
                    ("res_field", "=", "image_1920"),
                    ("res_model", "=", "product.template"),
                    ("res_id", "in", prod_ids),
                ],
                fields=["res_id"],
            )
            LOGGER.debug("Filtering %s products that already have an image", len(existing_images))
            existing_image_ids = [x["res_id"] for x in existing_images]
            prod_ids = [i for i in prod_ids if i not in existing_image_ids]

        if not prod_ids:
            return

        LOGGER.info("Getting %s Products from Odoo", len(prod_ids))
        for index, prod_id in enumerate(prod_ids, 1):
            prod = products_model.browse([prod_id])
            dcode = prod.default_code
            LOGGER.debug("Searching images for %s", dcode)
            match_images = [v[1] for v in images if v[0] == dcode]
            if match_images:
                selected_img = match_images[0]
                LOGGER.info(
                    "(%s/%s) Setting Product image for '%s' --> '%s'", index, len(prod_ids), dcode, selected_img
                )
                img_b64 = base64.b64encode(selected_img.open("rb").read()).decode("utf-8")
                prod.image_1920 = img_b64

    def _files_by_regex(self, root_folder: Path, regex_pattern: str) -> Dict[Match[str], Path]:
        """Recursively searches for images by a regex.

        Parameters
        ----------
        root_folder : Path
            path to recursively scan for
        regex_pattern : str
            regex pattern to match against

        Returns
        -------
        Dict[re.Match[str], Path]
            dict of regex match and file
        """
        LOGGER.info("Searching Product images in '%s' Regex: '%s'", root_folder, regex_pattern)
        regex = re.compile(regex_pattern)
        images = {f_match: f for f in root_folder.rglob("*") if (f_match := regex.match(f.name))}
        LOGGER.debug("Found %s images", len(images))
        return images

    def search_images_by_regex(self, image_path: Path, regex_pattern: str) -> List[Tuple[str, Path]]:
        """Search folder recursive for images matching a regex and then uploady matching ones to the odoo products.

        Parameters
        ----------
        image_path : Path
            folder to scan for images
        regex_pattern : str
            regex pattern with named group "default_code". (Group is matched against odoo "default_code")

        Returns
        -------
        List[tuple[str,Path]]
            mapping default_code to image file
        """
        images = self._files_by_regex(image_path, regex_pattern)

        return [(m.group("default_code"), f) for m, f in images.items()]
